/*************************************************************************
* (C) 2008 AITNET ltd - Sofia/Bulgaria - <misho@aitbg.com>
*  by Michael Pounov <misho@openbsd-bg.org>
*
* $Author: misho $
* $Id: aitsess.h,v 1.6 2014/02/08 22:06:16 misho Exp $
*
**************************************************************************
The ELWIX and AITNET software is distributed under the following
terms:

All of the documentation and software included in the ELWIX and AITNET
Releases is copyrighted by ELWIX - Sofia/Bulgaria <info@elwix.org>

Copyright 2004 - 2014
	by Michael Pounov <misho@elwix.org>.  All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:
1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. All advertising materials mentioning features or use of this software
   must display the following acknowledgement:
This product includes software developed by Michael Pounov <misho@elwix.org>
ELWIX - Embedded LightWeight unIX and its contributors.
4. Neither the name of AITNET nor the names of its contributors
   may be used to endorse or promote products derived from this software
   without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY AITNET AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
*/
#ifndef __AITSESS_H
#define __AITSESS_H


#include <pthread.h>
#include <semaphore.h>
#include <elwix.h>
#include <aitio.h>

#define SHARED_UNKNOWN	-1
#define SHARED_IPC	0
#define SHARED_MAP	1

#define SESS_OPT_SEED	-1

#define SESS_F_ADD	0x10000000
#define SESS_F_DEF	0x20000000

#define SESS_RET_NEW(x)	(((x) & SESS_F_ADD) == SESS_F_ADD)
#define SESS_RET_DEF(x)	(((x) & SESS_F_DEF) == SESS_F_DEF)
#define SESS_RET_VAL(x)	((x) & 0x0fffffff)

/* Shared memory session */

typedef struct tagSess {
	key_t	sess_key;
	char	sess_type;
	char	sess_zcpy;

	char	sess_name[STRSIZ];
	off_t	sess_size;
	void	*sess_addr;
	off_t	sess_offset;

	union {
		int	shmid;
		int	fd;
	} sess_mem;
	union {
		int	semid;
		sem_t	*sid;
	} sess_id;

	/* Session callbacks */
	struct {
		int (*create)(int, long, void *, ...);
		void (*destroy)(void *);
		void *(*attach)(void *, void *);
		void (*detach)(void *);
		void (*notSem)(void *);
		int (*isSemOK)(void *);
		int (*incSem)(void *);
		int (*decSem)(void *);
	} sess;
} sess_t;

/* --------------------------------------------------------- */

#define ALLOC_MEMORY(sd, siz, s, ...)	(assert((s) && (s)->sess.create), \
						(s)->sess.create((sd), (siz), (s), ## __VA_ARGS__))
#define ATTACH_MEMORY(s)		(assert((s) && (s)->sess.attach), (s)->sess.attach((s), NULL))
#define DETACH_MEMORY(s)		(assert((s) && (s)->sess.detach), (s)->sess.detach((s)))
#define FREE_MEMORY(s)			(assert((s) && (s)->sess.destroy), (s)->sess.destroy((s)))

#define IS_SEMOK(s)			(assert((s) && (s)->sess.isSemOK), (s)->sess.isSemOK((s)))
#define INC_SEM(s)			(assert((s) && (s)->sess.incSem), (s)->sess.incSem((s)))
#define DEC_SEM(s)			(assert((s) && (s)->sess.decSem), (s)->sess.decSem((s)))
#define NOT_SEM(s)			(assert((s) && (s)->sess.notSem), (s)->sess.notSem((s)))

/* --------------------------------------------------------- */


/*
 * sess_initSession() - Initializing session structure, 
 * 			if session file not exists creating with specified tech
 *
 * @id = Technology using in session. SHARED_IPC IPC tech or SHARED_MAP BSD MemoryMap tech
 * @csFName = Session filename for build key and identified
 * @Sess = Session item, if =NULL allocate memory for session after use must be free!
 * return: 0 OK new key created, -1 error: no memory or file not created, 1 OK key finded
 */
int sess_initSession(int id, const char *csFName, sess_t ** __restrict Sess);
/*
 * sess_freeSession() - Free allocated memory for session item and delete session file if present name
 *
 * @Sess = Session item
 * return: none
 */
void sess_freeSession(sess_t ** __restrict Sess);

/*
 * map_createSession() - MMAP Created session and allocated resources
 *
 * @nSeed = Seed for securing key, if =SESS_OPT_SEED must add ready for use key
 * @nSize = Allocated shared memory size in bytes
 * @Sess = Session item
 * @... = If nSeed == -1 add ready for use key value
 * return: 0 Ok successful, -1 error: not allocated resources
 */
int map_createSession(int nSeed, long nSize, sess_t * __restrict Sess, ...);
/*
 * map_destroySession() - MMAP free shared resources
 *
 * @Sess = Session item
 * return: none
 */
void map_destroySession(sess_t * __restrict Sess);

/*
 * ipc_createSession() - IPC Created session and allocated resources
 *
 * @nSeed = Seed for securing key, if =SESS_OPT_SEED must add ready for use key
 * @nSize = Allocated shared memory size in bytes
 * @Sess = Session item
 * @... = If nSeed == -1 add ready for use key value
 * return: 0 Ok successful, -1 error: not allocated resources
 */
int ipc_createSession(int nSeed, long nSize, sess_t * __restrict Sess, ...);
/*
 * ipc_destroySession() - IPC free shared resources
 *
 * @Sess = Session item
 * return: none
 */
void ipc_destroySession(sess_t * __restrict Sess);

/*
 * map_attachSession() - MMAP Attach to shared memory & return begin address
 *
 * @s = Session item
 * @procMem = Custom start address (optionl) *default must be 0*
 * return: NULL failed attach, !=NULL begin address of memory
 */
void *map_attachSession(sess_t * __restrict s, void *procMem);
/*
 * map_detachSession() - MMAP Detach from shared memory
 *
 * @s = Session item
 * return: none
 */
void map_detachSession(sess_t * __restrict s);

/*
 * ipc_attachSession() - IPC Attach to shared memory & return begin address
 *
 * @s = Session item
 * @procMem = Custom start address (optionl) *default must be 0*
 * return: NULL failed attach, !=NULL begin address of memory
 */
void *ipc_attachSession(sess_t * __restrict s, void *procMem);
/*
 * ipc_detachSession() - IPC Detach from shared memory
 *
 * @s = Session item
 * return: none
 */
void ipc_detachSession(sess_t * __restrict s);

/*
 * sess_isAttached() - Check for attached shared memory
 *
 * @s = Session item
 * return: -1 null session item, 0 not attached, 1 attached memory
 */
int sess_isAttached(sess_t * __restrict s);


/*
 * map_notSemaphore() - MMAP negative block if semaphore isn`t signaled
 *
 * @s = Session item
 * return: none
 */
void map_notSemaphore(sess_t * __restrict s);
/*
 * map_isSemaphoreOK() - MMAP Check semaphore
 *
 * @s = Session item
 * return: -1 error: can`t return semaphore, 0 = false, 1 = true
 */
int map_isSemaphoreOK(sess_t * __restrict s);
/*
 * map_incSemaphore() - MMAP unblock semaphore, increment semaphore
 *
 * @s = Session item
 * return: 0 Ok, -1 error: can`t increment 
 */
int map_incSemaphore(sess_t * __restrict s);
/*
 * map_decSemaphore() - MMAP block semaphore, decrement semaphore
 *
 * @s = Session item
 * return: 0 Ok, -1 error: can`t decrement 
 */
int map_decSemaphore(sess_t * __restrict s);

/*
 * ipc_notSemaphore() - IPC negative block if semaphore isn`t signaled
 *
 * @s = Session item
 * return: none
 */
void ipc_notSemaphore(sess_t * __restrict s);
/*
 * ipc_isSemaphoreOK() - IPC Check semaphore
 *
 * @s = Session item
 * return: -1 error: can`t return semaphore, 0 = false, 1 = true
 */
int ipc_isSemaphoreOK(sess_t * __restrict s);
/*
 * ipc_incSemaphore() - IPC unblock semaphore, increment semaphore
 *
 * @s = Session item
 * return: 0 Ok, -1 error: can`t increment 
 */
int ipc_incSemaphore(sess_t * __restrict s);
/*
 * ipc_decSemaphore() - IPC block semaphore, decrement semaphore
 *
 * @s = Session item
 * return: 0 Ok, -1 error: can`t decrement 
 */
int ipc_decSemaphore(sess_t * __restrict s);


/*
 * sess_GetValues() - Get all values from session shared memory
 *
 * @s = Session item
 * @Vals = Return array strings
 * return: -1 error: in parameter, !=-1 count of returned strings in Vals 
 * 		(must call ait_freeVars() after use!)
 */
int sess_GetValues(sess_t * __restrict s, array_t ** __restrict Vals);
/*
 * sess_GetValue() - Get value from session shared memory from attribute
 *
 * @s = Session item
 * @csAttr = Attribute for search
 * @v = Return string value and after use must call ait_freeVar(), also may be =NULL
 * return: 0 not found, -1 error: in parameter, >0 get position, 
 * 	if define item merged with IS_DEF
 */
int sess_GetValue(sess_t * __restrict s, const char *csAttr, ait_val_t ** __restrict v);
/*
 * sess_DelValue() - Delete item from session shared memory
 *
 * @s = Session item
 * @csAttr = Attribute for erasing
 * return: -1 error: in parameter or !=-1 deleted items
 */
int sess_DelValue(sess_t * __restrict s, const char *csAttr);
/*
 * sess_SetValue() - Set item into session shared memory or update if exists
 *
 * @s = Session item
 * @csAttr = Attribute
 * @psVal = Value
 * return: 0 nothing, -1 error: in parameter, 
 	>0 set position, if added new item merged with SESS_F_ADD and 
	if just define item merged with SESS_F_DEF
 */
int sess_SetValue(sess_t * __restrict s, const char *csAttr, const char *psVal);

/*
 * sess_prepareSession() - Attach to shared memory and de-marshaling data
 *
 * @s = Session
 * @useDirect = Use direct shared memory if !=0 or snapshot of data to array
 * return: NULL error or no data, !=NULL array with variables, 
 *		after use must free resources with sess_doneSession()
 */
array_t *sess_prepareSession(sess_t * __restrict s, char useDirect);
/*
 * sess_doneSession() - Free resources allocated with sess_prepareSession()
 *
 * @s = Session
 * @arr = Array with variables for free
 * return: none
 */
void sess_doneSession(sess_t * __restrict s, array_t ** __restrict arr);
/*
 * sess_commitSession() - Commit data to shared memory
 *
 * @s = Session
 * @arr = Array with variables for save
 * return -1 error or !=-1 size of stored variables into shared memory
 */
int sess_commitSession(sess_t * __restrict s, array_t * __restrict arr);


#endif
